/****************************************************************************

    PROGRAM: wintel.c

    PURPOSE: Windows networking kernel - Telnet

    FUNCTIONS:

    WinMain() - calls initialization function, processes message loop
    InitApplication() - initializes window data and registers window
    InitInstance() - saves instance handle and creates main window
    MainWndProc() - processes messages
    About() - processes messages for "About" dialog box

    COMMENTS:

        Windows can have several copies of your application running at the
        same time.  The variable hInst keeps track of which instance this
        application is so that processing will be to the correct window.

****************************************************************************/

#include <windows.h>            /* required for all Windows applications */
#include <string.h>
#include "wintel.h"            /* specific to this program          */

HANDLE hInst;   /* current instance      */
HWND hWnd;      /* Main window handle.   */
CONFIG FAR *tmpConfig;
GLOBALHANDLE hGlobalMem,hTitleMem;
CONNECTION far *con=NULL;
char hostdata[MAXGETHOSTSTRUCT];
HGLOBAL ghCon;
SCREEN far *fpScr;
int debug=1;
char strTmp[512];
BOOL bAutoConnection=FALSE; 
char szAutoHostName[64];

/****************************************************************************

    FUNCTION: WinMain(HANDLE, HANDLE, LPSTR, int)

    PURPOSE: calls initialization function, processes message loop

    COMMENTS:

        Windows recognizes this function by name as the initial entry point 
        for the program.  This function calls the application initialization 
        routine, if no other instance of the program is running, and always 
        calls the instance initialization routine.  It then executes a message 
        retrieval and dispatch loop that is the top-level control structure 
        for the remainder of execution.  The loop is terminated when a WM_QUIT 
        message is received, at which time this function exits the application 
        instance by returning the value passed by PostQuitMessage(). 

        If this function must abort before entering the message loop, it 
        returns the conventional value NULL.  

****************************************************************************/

int PASCAL WinMain(hInstance, hPrevInstance, lpCmdLine, nCmdShow)
HANDLE hInstance;                /* current instance             */
HANDLE hPrevInstance;            /* previous instance            */
LPSTR lpCmdLine;                 /* command line                 */
int nCmdShow;                    /* show-window type (open/icon) */
{
    MSG msg;                     /* message                      */

    if (!hPrevInstance)
    	if (!InitApplication(hInstance)) /* Initialize shared things */
        	return (FALSE);      /* Exits if unable to initialize    */

    /* Perform initializations that apply to a specific instance */

	if (lpCmdLine[0]) {
		bAutoConnection=TRUE;
		lstrcpy((char far *)szAutoHostName,lpCmdLine);
	}
	else bAutoConnection=FALSE;
	
    if (!InitInstance(hInstance, nCmdShow))
        return (FALSE);

    /* Acquire and dispatch messages until a WM_QUIT message is received. */

    while (GetMessage(&msg, NULL, NULL, NULL)) {
        TranslateMessage(&msg); /* Translates virtual key codes  */
        DispatchMessage(&msg);  /* Dispatches message to window  */
    }
    return (msg.wParam);    /* Returns the value from PostQuitMessage */
}


/****************************************************************************

    FUNCTION: InitApplication(HANDLE)

    PURPOSE: Initializes window data and registers window class

    COMMENTS:

        This function is called at initialization time only if no other 
        instances of the application are running.  This function performs 
        initialization tasks that can be done once for any number of running 
        instances.  

        In this case, we initialize a window class by filling out a data 
        structure of type WNDCLASS and calling the Windows RegisterClass() 
        function.  Since all instances of this application use the same window 
        class, we only need to do this when the first instance is initialized.  


****************************************************************************/

BOOL InitApplication(hInstance)
HANDLE hInstance;                  /* current instance       */
{
    WNDCLASS  wc;

    /* Fill in window class structure with parameters that describe the       */
    /* main window.                                                           */

    wc.style = CS_HREDRAW|CS_VREDRAW;                    /* Class style(s).                    */
    wc.lpfnWndProc = MainWndProc;       /* Function to retrieve messages for  */
                                        /* windows of this class.             */
    wc.cbClsExtra = 0;                  /* No per-class extra data.           */
    wc.cbWndExtra = 0;                  /* No per-window extra data.          */
    wc.hInstance = hInstance;           /* Application that owns the class.   */
    wc.hIcon = NULL; //LoadIcon(hInstance, "NCSA");
    wc.hCursor = NULL; //Cursor(NULL, IDC_ARROW);
    wc.hbrBackground = NULL; //GetStockObject(WHITE_BRUSH); 
    wc.lpszMenuName =  NULL;   /* Name of menu resource in .RC file. */
    wc.lpszClassName = "wintelWClass"; /* Name used in call to CreateWindow. */


    /* Register the window class and return success/failure code. */
    return (RegisterClass(&wc));
}


/****************************************************************************

    FUNCTION:  InitInstance(HANDLE, int)

    PURPOSE:  Saves instance handle and creates main window

    COMMENTS:

        This function is called at initialization time for every instance of 
        this application.  This function performs initialization tasks that 
        cannot be shared by multiple instances.  

        In this case, we save the instance handle in a static variable and 
        create and display the main program window.  
        
****************************************************************************/
BOOL InitInstance(hInstance, nCmdShow)
    HANDLE          hInstance;          /* Current instance identifier.       */
    int             nCmdShow;           /* Param for first ShowWindow() call. */
{
    int xScreen=0, yScreen=0;

    /* Save the instance handle in static variable, which will be used in  */
    /* many subsequence calls from this application to Windows.            */

    hInst = hInstance;

    /* Create a main window for this application instance.  */
    hWnd = CreateWindow(
        "wintelWClass",                /* See RegisterClass() call.          */
        "TCPWin",                       /* Text for window title bar.         */
        WS_SYSMENU,                     /* Window style.                      */
        xScreen/3,                      /* Default horizontal position.       */
        yScreen/3,                      /* Default vertical position.         */
        xScreen/3,                      /* Default width.                     */
        yScreen/3,                      /* Default height.                    */
        NULL,                           /* Overlapped windows have no parent. */
        NULL,                           /* Use the window class menu.         */
        hInstance,                      /* This instance owns this window.    */
        NULL                            /* Pointer not needed.                */
    );
    if (!hWnd)
        return (FALSE);
        
//    ShowWindow(hWnd, SW_SHOW);  /* Show the window   */
//    UpdateWindow(hWnd);          /* Sends WM_PAINT message      */        

    {
        WSADATA wsaData;
        if (WSAStartup(0x0101,&wsaData)!=0) {   /* Initialize the network */
	        MessageBox(NULL,"Couldn't initialize Winsock!",NULL,MB_OK|MB_ICONEXCLAMATION);
	        return FALSE;
		}        	
    }

	if (!OpenTelnetConnection()) {
		WSACleanup();
		return(FALSE);
	}		

    return (TRUE);
}

/****************************************************************************

    FUNCTION: MainWndProc(HWND, UINT, WPARAM, LPARAM)

    PURPOSE:  Processes messages

    MESSAGES:

    WM_COMMAND    - application menu (About dialog box)
    WM_DESTROY    - destroy window

****************************************************************************/

long FAR PASCAL MainWndProc(HWND hWnd, UINT message, WPARAM wParam, LPARAM lParam) {
    HSCREEN hgScr;
    SCREEN far *fpScr;
    HGLOBAL hBuffer;
    LPSTR lpBuffer; 
    char c;
    int iEvent,namelen,cnt,ret;
    char buf[1024];
    struct sockaddr_in name;
	struct sockaddr_in         remote_addr;
    struct hostent          far *remote_host;
            
    
    switch (message) {
	    case WM_MYSCREENCHANGEBKSP:
	        if (!con) break;
	        con->backspace=wParam;
	        if (con->backspace==VK_BACK) {
	            con->ctrl_backspace=0x7f;
	        } else {    
	            con->ctrl_backspace=VK_BACK;
	        }    
	        break;
	
		
	    case WM_MYSCREENCHAR:
	        if (!con) break;
	        if (wParam==VK_BACK)
	            wParam=con->backspace;
	          else
	            if (wParam==0x7f)
	                wParam=con->ctrl_backspace;
	        send(con->socket, (char FAR *)&wParam,1,NULL);
	        break;
	    
	    case WM_MYSYSCHAR:
	        if (!con) break;
	        c= (char) wParam;
	        switch (c) {
	            case 'f':
	                getsockname(con->socket,(struct sockaddr far *)&name,(int FAR *)&namelen);
	                wsprintf(buf,"ftp %d.%d.%d.%d\n",
	                    name.sin_addr.S_un.S_un_b.s_b1,
	                    name.sin_addr.S_un.S_un_b.s_b2,
	                    name.sin_addr.S_un.S_un_b.s_b3,
	                    name.sin_addr.S_un.S_un_b.s_b4);
	                send(con->socket,buf,lstrlen(buf),NULL);
	                break;                
	            case 'x':
	                hgScr=con->hScreen;
	                fpScr=(SCREEN far *)GlobalLock(hgScr);
	                if (fpScr == NULL) break;
	                PostMessage(fpScr->hWnd,WM_CLOSE,NULL,NULL);
	                break;
	                                
	        }                
	        break;
	            
	    case WM_MYSCREENBLOCK:
	        if (!con) break;
	        hBuffer= (HGLOBAL) wParam;
	        lpBuffer= GlobalLock(hBuffer);
	        send(con->socket,lpBuffer,lstrlen(lpBuffer),NULL);
	        GlobalUnlock(hBuffer);
	        break;
	
	    case WM_MYSCREENCLOSE:
	        if (!con) break;
	        closesocket(con->socket);
	        GlobalUnlock(ghCon);
	        GlobalFree(ghCon);
	        break;        
	
	    case WM_QUERYOPEN:
	        return(0);
	        break;
	
	    case WM_DESTROY:          /* message: window being destroyed */
	        closesocket(con->socket);
	        GlobalUnlock(ghCon);
	        GlobalFree(ghCon); 
	        WSACleanup();
	        PostQuitMessage(0);
	        break;
	
	    case WM_NETWORKEVENT: 
			iEvent=WSAGETSELECTEVENT(lParam);
			switch (iEvent) {    	
				case FD_READ:
	                cnt=recv(con->socket,buf,1024,NULL);
	                buf[cnt]=0;
	                parse((CONNECTION far *)con,(unsigned char far *)buf,(int)lstrlen(buf));
	                ScreenEm(buf,cnt,con->hScreen);
	                break;
	            case FD_CLOSE:
	            	closesocket(con->socket);
			        GlobalUnlock(ghCon);
			        GlobalFree(ghCon); 
			        WSACleanup();
			        PostQuitMessage(0);
					break;
				case FD_CONNECT:
					ret = WSAGETSELECTERROR(lParam);
					if (ret) {
						wsprintf(buf,"Error %d on Connect",ret);
		        		MessageBox(NULL,buf,NULL,MB_OK|MB_ICONEXCLAMATION);
		            	closesocket(con->socket);
				        GlobalUnlock(ghCon);
				        GlobalFree(ghCon); 
				        WSACleanup();
				        PostQuitMessage(0);
				        break;
		        	}
				    //start_negotiation(con->socket);		        	
					break;		
  	        }
  	        break;  	        

		case WM_HOSTNAMEFOUND:
			ret = WSAGETASYNCERROR(lParam);
			if (ret) {
				wsprintf(buf,"Error %d on GetHostbyName",ret);
        		MessageBox(NULL,buf,NULL,MB_OK|MB_ICONEXCLAMATION);
            	closesocket(con->socket);
		        GlobalUnlock(ghCon);
		        GlobalFree(ghCon); 
		        WSACleanup();
		        PostQuitMessage(0);
        		break;
        	}
        	remote_host= (struct hostent far *) hostdata;
		    remote_addr.sin_family= AF_INET;    
		    remote_addr.sin_addr.S_un.S_un_b.s_b1= remote_host->h_addr[0];
		    remote_addr.sin_addr.S_un.S_un_b.s_b2= remote_host->h_addr[1];
		    remote_addr.sin_addr.S_un.S_un_b.s_b3= remote_host->h_addr[2];
		    remote_addr.sin_addr.S_un.S_un_b.s_b4= remote_host->h_addr[3];
		    remote_addr.sin_port= htons(23);
		    
		    connect(con->socket,(struct sockaddr far*) &remote_addr, 
		    		sizeof(struct sockaddr));    
			break;
			
	    default:              /* Passes it on if unproccessed    */
	        return (DefWindowProc(hWnd, message, wParam, lParam));
    }
    return (NULL);
}

int OpenTelnetConnection(void) {
    int nReturn,ret;
    struct sockaddr_in         sockaddr;
    char buf[128],tmpName[128],far * tmpCommaLoc;
    int iHostNum=0;
        
    hGlobalMem= GlobalAlloc(GPTR,sizeof(CONFIG));
    tmpConfig= (CONFIG FAR *) GlobalLock(hGlobalMem);
    
    if (bAutoConnection) {
		hTitleMem= GlobalAlloc(GPTR,lstrlen(szAutoHostName));
   		tmpConfig->title= (char far *) GlobalLock(hTitleMem);            
		lstrcpy(tmpConfig->title,(char far *)szAutoHostName);
    } else {
    	nReturn=DoDialog("OPENTELNETDLG",OpenTelnetDlg);
    	if (nReturn==FALSE) return(FALSE);
	}
	tmpConfig->width=DEF_WIDTH;
    tmpConfig->height=DEF_HEIGHT;                            
                     	
    con=(CONNECTION far *)GetNewConnection();
    if (con==NULL) return(0);

    for(;;) {
        char tmpBuf[80];
        
        wsprintf(tmpBuf,"host%d",iHostNum++);
        GetPrivateProfileString("hosts",tmpBuf,"",tmpName,128,"wintel.ini");
        tmpCommaLoc=_fstrchr(tmpName,',');  
        if (tmpCommaLoc) *tmpCommaLoc='\0';
        if (tmpName[0])
            if (!lstrcmp(tmpName,tmpConfig->title)) break;
            else continue;
        else break; 
    }

    if (tmpCommaLoc) tmpCommaLoc++;
    if (tmpName[0]) 
        if (!lstrcmp(tmpCommaLoc,"backspace")) con->backspace=VK_BACK;
            else con->backspace=0x7f; 
    else con->backspace=VK_BACK;        
    
    if (con->backspace==0x7f) {
        con->ctrl_backspace=0x08;
        tmpConfig->backspace=FALSE;
    }
    else {
        tmpConfig->backspace=TRUE;
        con->ctrl_backspace=0x7f;
    }
    tmpConfig->hwndTel=hWnd;
    con->hScreen=(HSCREEN)InitNewScreen((CONFIG FAR *)tmpConfig);
    if (!(con->hScreen)) {
    	OutputDebugString("Failed to initialize new screen! \r\n");
        GlobalUnlock(con->hScreen);
        GlobalUnlock(ghCon);
        GlobalFree(ghCon);
        GlobalUnlock(hGlobalMem); 
        GlobalFree(hGlobalMem);
        return(-1);
    }
    ret=(SOCKET)socket(AF_INET,SOCK_STREAM,IPPROTO_TCP);
    
    if (ret==SOCKET_ERROR) {
        wsprintf(buf,"Socket error on socket = %d!",WSAGetLastError());
        MessageBox(NULL,buf,NULL,MB_OK|MB_ICONEXCLAMATION);
        fpScr= (SCREEN far *) GlobalLock(con->hScreen);
        if (fpScr!=NULL)
        	DestroyWindow(fpScr->hWnd);
        GlobalUnlock(con->hScreen);
        GlobalUnlock(ghCon);
        GlobalFree(ghCon);
        GlobalUnlock(hGlobalMem);
        GlobalFree(hGlobalMem);
        return(-1);
    } 
    
    con->socket=ret;
    
    sockaddr.sin_family = AF_INET ;  
    sockaddr.sin_addr.s_addr = htonl( INADDR_ANY );
    sockaddr.sin_port = htons( 0 ) ;
    
    ret=bind(con->socket,(struct sockaddr far *)&sockaddr,(int)sizeof(struct sockaddr_in));
    
    if (ret==SOCKET_ERROR) {
        wsprintf(buf,"Socket error on bind!");
        MessageBox(NULL,buf,NULL,MB_OK|MB_ICONEXCLAMATION);
        fpScr= (SCREEN far *) GlobalLock(con->hScreen);
        if (fpScr!=NULL)
        	DestroyWindow(fpScr->hWnd);
        GlobalUnlock(con->hScreen);
        GlobalUnlock(ghCon);
        GlobalFree(ghCon);
        GlobalUnlock(hGlobalMem);
        GlobalFree(hGlobalMem);
        return(-1);
    }
    WSAAsyncSelect(con->socket,hWnd,WM_NETWORKEVENT,FD_READ|FD_CLOSE|FD_CONNECT);
    
   	WSAAsyncGetHostByName(hWnd,WM_HOSTNAMEFOUND,tmpConfig->title,hostdata,MAXGETHOSTSTRUCT); 
    GlobalUnlock(con->hScreen);
//    GlobalUnlock(ghCon);
//    GlobalFree(ghCon);
    GlobalUnlock(hGlobalMem);
    GlobalFree(hGlobalMem);

#ifdef OLD_WAY
    remote_host = gethostbyname(tmpConfig->title);
    if (!remote_host) {
        wsprintf(buf,"Unable to lookup %s!",tmpConfig->title);
        MessageBox(NULL,buf,NULL,MB_OK|MB_ICONEXCLAMATION);
        fpScr= (SCREEN far *) GlobalLock(con->hScreen);
        if (fpScr!=NULL)
        	DestroyWindow(fpScr->hWnd);
        GlobalUnlock(con->hScreen);
        GlobalUnlock(ghCon);
        GlobalFree(ghCon);
        GlobalUnlock(hGlobalMem);
        GlobalFree(hGlobalMem);
        return(-1);
    }        

    remote_addr.sin_family= AF_INET;    
    remote_addr.sin_addr.S_un.S_un_b.s_b1= remote_host->h_addr[0];
    remote_addr.sin_addr.S_un.S_un_b.s_b2= remote_host->h_addr[1];
    remote_addr.sin_addr.S_un.S_un_b.s_b3= remote_host->h_addr[2];
    remote_addr.sin_addr.S_un.S_un_b.s_b4= remote_host->h_addr[3];
    remote_addr.sin_port= htons(23);
    
    if(connect(con->socket,(struct sockaddr far*) &remote_addr, sizeof(struct sockaddr))){
        MessageBox(NULL,"Connect error!",NULL,MB_OK|MB_ICONEXCLAMATION);
        fpScr= (SCREEN far *) GlobalLock(con->hScreen);
        if (fpScr!=NULL)
        	DestroyWindow(fpScr->hWnd);
        GlobalUnlock(con->hScreen);
        GlobalUnlock(ghCon);
        GlobalFree(ghCon);
        GlobalUnlock(hGlobalMem);
        GlobalFree(hGlobalMem);
        return(-1);
    }    
	    
    start_negotiation(con->socket);
#endif    
    return(1);
}

CONNECTION far *GetNewConnection(void) {
    CONNECTION far *fpCon;

    ghCon = GlobalAlloc(GHND,sizeof(CONNECTION));
    if (ghCon==NULL) return NULL;
    fpCon = (CONNECTION far *)GlobalLock(ghCon);
    if (fpCon==NULL) return NULL;
    fpCon->backspace=TRUE;
    return(fpCon);
}
   
/****************************************************************************/
int NEAR DoDialog(char *szDialog, FARPROC lpfnDlgProc) {
    int nReturn;
    
    lpfnDlgProc= MakeProcInstance(lpfnDlgProc,hInst);
    if (lpfnDlgProc==NULL)
        MessageBox(hWnd,"Couldn't make procedure instance",NULL,MB_OK);    
    
    nReturn= DialogBox(hInst,szDialog,hWnd,lpfnDlgProc);
    FreeProcInstance(lpfnDlgProc);
    return (nReturn);
}    

/****************************************************************************

    FUNCTION: OpenTelnetDlg(HWND, unsigned, WORD, LONG)

    PURPOSE:  Processes messages for "Open New Telnet Connection" dialog box

    MESSAGES:

    WM_INITDIALOG - initialize dialog box
    WM_COMMAND    - Input received

****************************************************************************/

BOOL FAR PASCAL OpenTelnetDlg(HWND hDlg, WORD message, WORD wParam, LONG lParam) {
    char szConnectName[256];
    HDC  hDC;
    int  xExt,yExt;
    DWORD Ext;
    HWND hEdit;
    int  nLen;
    int iHostNum=1;
    char tmpName[128],tmpBuf[80];
    
    switch (message) {
        case WM_INITDIALOG:        /* message: initialize dialog box */
            hDC=GetDC(hDlg);
            Ext=GetDialogBaseUnits();
            xExt=(190 *LOWORD(Ext)) /4 ;
            yExt=(72 * HIWORD(Ext)) /8 ;
            GetPrivateProfileString("hosts","host0","",tmpName,128,"wintel.ini");
            if (tmpName[0]) {
            	*(_fstrchr(tmpName,','))='\0';  
            	SetDlgItemText(hDlg,TEL_CONNECT_NAME,tmpName);
            }	
            hEdit=GetWindow(GetDlgItem(hDlg,TEL_CONNECT_NAME),GW_CHILD);
            for(;;) {
                char far *tmpCommaLoc;
                wsprintf(tmpBuf,"host%d",iHostNum++);
                GetPrivateProfileString("hosts",tmpBuf,"",tmpName,128,"wintel.ini");
                tmpCommaLoc=_fstrchr(tmpName,',');  
                if (tmpCommaLoc) *tmpCommaLoc='\0';
                if (tmpName[0])
                    SendDlgItemMessage(hDlg,TEL_CONNECT_NAME,CB_ADDSTRING,0,(LPARAM)((LPSTR)tmpName));
                else break; 
            }
            SetWindowPos(hDlg,NULL,(GetSystemMetrics(SM_CXSCREEN)/2)-(xExt/2),
                (GetSystemMetrics(SM_CYSCREEN)/2)-(yExt/2),0,0,
                SWP_NOSIZE|SWP_NOZORDER|SWP_SHOWWINDOW);
            ReleaseDC(hDlg,hDC);      
            SendMessage(hEdit,WM_USER+1,NULL,NULL);
            SendMessage(hDlg,WM_SETFOCUS,NULL,NULL);
            return (TRUE);

        case WM_COMMAND:              /* message: received a command */
            switch (wParam) {
                case TEL_CANCEL:
                     EndDialog(hDlg, FALSE);        /* Exits the dialog box        */
                     break;
                 
                case TEL_OK:
                     GetDlgItemText(hDlg,TEL_CONNECT_NAME,szConnectName,256);
                     nLen= lstrlen(szConnectName);
                     if (nLen==0) { 
                        MessageBox(hDlg,"You must enter a session name!",NULL,MB_OK);
                        break;
                        }
                     hTitleMem= GlobalAlloc(GPTR,nLen);
                     tmpConfig->title= (char far *) GlobalLock(hTitleMem);
                     lstrcpy(tmpConfig->title,szConnectName);
                     EndDialog(hDlg,TRUE);
                     break;
            }                
            return (FALSE);                    
    }
    return (FALSE);               /* Didn't process a message    */
}

